"""
Test the core functionality.

Contains a converted version of all of the doctests.

"""
# test utilities
import unittest
import weakref, gc

# tested module
from obsub import event


class NewStyle(object):
    def __init__(self):
        self.count = 0
    @event
    def emit(self, first, second):
        self.count += 1

class OldStyle:
    def __init__(self):
        self.count = 0
    @event
    def emit(self, first, second):
        self.count += 1

class Observer(object):
    def __init__(self, call_stack):
        self.call_stack = call_stack
    def __call__(self, source, first, second):
        diff = (sum(1 for call in self.call_stack if call[2] == source)
                - source.count) + 1
        self.call_stack.append((self, diff, source, first, second))

class TestCore(unittest.TestCase):
    """Test the obsub module for new style classes."""
    cls = NewStyle

    def setUp(self):
        self.call_stack = []
        self.maxDiff = None

    def observer(self, instance):
        observer = Observer(self.call_stack)
        instance.emit += observer
        return observer

    def check_stack(self, expected):
        self.assertEqual(expected, self.call_stack)

    def test_single_handler(self):
        """A single handler is invoked correctly."""
        src = self.cls()
        obs = self.observer(src)
        src.emit("Hello", "World")
        self.check_stack([(obs, 0, src, "Hello", "World")])

    def test_remove_handler(self):
        """Removal of event handlers works correctly."""
        src = self.cls()
        obs = self.observer(src)
        src.emit -= obs
        src.emit("something", "arbitrary")
        self.check_stack([])

    def test_multiple_handlers(self):
        """Multiple handlers are invoked in correct order."""
        src = self.cls()
        obs = [self.observer(src) for i in range(10)]
        src.emit("Hello", "World")
        self.check_stack([(obs[i], i, src, "Hello", "World")
                          for i in range(10)])

    def test_multiple_sources(self):
        """Multiple instances of the event source class can coexist."""
        src = [self.cls() for i in range(10)]
        obs = [self.observer(s) for s in src]
        for s in src:
            s.emit("Hello", "World")
        self.check_stack([(obs[i], 0, src[i], "Hello", "World")
                          for i in range(10)])

    def test_keyword_arguments(self):
        """Keyword arguments can be used."""
        src = self.cls()
        obs = self.observer(src)
        src.emit(second="World", first="Hello")
        self.check_stack([(obs, 0, src, "Hello", "World")])

    def test_wrong_signature(self):
        """Any incorrect call signature raises a TypeError."""
        src = self.cls()
        self.assertRaises(TypeError, src.emit, "Hello")
        self.assertRaises(TypeError, src.emit, first="Hello", third="!")
        self.assertRaises(TypeError, src.emit, second="World")
        obs = self.observer(src)
        self.assertRaises(TypeError, src.emit, forth=":)")
        self.check_stack([])

    def test_class_based_access(self):
        """The emitter function can be invoked via its class."""
        src = self.cls()
        obs = self.observer(src)
        self.cls.emit(src, "Hello", "World")
        self.check_stack([(obs, 0, src, "Hello", "World")])

    def test_keep_alive(self):
        """A reference to a bound event method keeps its owner instance alive."""
        src = self.cls()
        obs = self.observer(src)
        emit = src.emit

        # remove original reference:
        wr = weakref.ref(src)
        del src
        gc.collect()

        # the emitter function is still functioning:
        assert wr() is not None
        emit("Hello", "World")
        self.check_stack([(obs, 0, wr(), "Hello", "World")])

        # removing the bound method makes the instance disappear:
        del emit
        self.call_stack[0] = None
        gc.collect()
        assert wr() is None

# ERRORS:
class TestCoreOldStyle(TestCore):
    cls = OldStyle




'''
Test that the decorator correctly preserves the signature. Currently, this
is only possible for python3.
'''

from obsub import event

try:
    from inspect import signature
except ImportError:
    raise NotImplementedError


referenced = {}
def on_blubb(
        self:0,
        obscure:"NOTE: default argument is a by reference: "=referenced,
        *,
        with_kwonlyarg:int,
        **kwargs):
    return obscure

class A(object):
    on_blubb = event(on_blubb)


def test_signature():
    # Define a test class and an event handler
    a = A()

    # signature is preserved: (!!)
    assert signature(a.on_blubb) == signature(on_blubb)

    # NOTE: we even got the exact object as default parameter, not only an
    # exact copy:
    assert a.on_blubb(with_kwonlyarg="xyz") is referenced




'''
Test that the decorator correctly preserves the signature. Currently, this
is only possible for python3.
'''

import sys

try:
    from test.py3.test_signature import *
except ImportError:
    __test__ = False




'''
These tests that the event handling mechanism does not produce memory leaks.
This could in principle happen, since it introduces a cyclic dependency that
might prevent garbage collection.

'''

import weakref
import gc

from obsub import event


def test_memory_leak():

    # Define a test class and an event handler
    class A(object):
        @event
        def on_blubb(self):
            pass

    def handler(self):
        pass

    # Instantiate and attach event handler
    a = A()
    a.on_blubb += handler

    # Weak reference for testing
    wr = weakref.ref(a)

    # At first, weak reference exists
    assert wr() is not None

    # (implicitly) delete the A-instance by reassigning the only hard-ref.
    # This is equivalent to `del a` but serves the purpose to demonstrate
    # that there are very subtle ways to delete an instance:
    a = None

    # Trigger the garbage collection manually
    gc.collect()

    # after deletion it should be dead
    assert wr() is None


def test_object_stays_alive_during_handler_execution():

    # Define a test class and event handlers
    class A(object):
        @event
        def on_blubb(self):
            pass
        deleted = False

    class B(object):
        def __init__(self, a):
            # capture the only hard-ref on the A-instance:
            self.a = a
            self.a.on_blubb += self.handler

        def handler(self, a):
            # delete the A-instance
            del self.a
            a.deleted = True
            del a
            gc.collect()

    def handler(a):
        # We want a valid reference to the handled object, ...
        assert a is not None
        # ..., even if the deletion handler has already been called:
        assert a.deleted

    # Instantiate and attach event handler
    b = B(A())
    b.a.on_blubb += handler

    wr = weakref.ref(b.a)
    b.a.on_blubb()

    # Trigger the garbage collection manually
    gc.collect()

    # make sure, b.a has been deleted after event handling
    assert wr() is None



#!/usr/bin/env python

import logging
from setuptools import setup


# Create a front page for PyPI:
long_description = None
try:
    long_description = open('README.rst').read()
    long_description += '\n' + open('CHANGELOG.rst').read()
except IOError:
    # some file is not available
    # just use what we got so far but issue a warning
    logging.warn('documentation files missing')


setup(
    name='obsub',
    version='0.2.1',
    description='Implementation of the observer pattern via a decorator',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    author='Emilia Bopp',
    author_email='contact@ebopp.de',
    url='https://github.com/milibopp/obsub',
    py_modules=['obsub',],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development',
    ],
    license='Public Domain',
    tests_require='nose',
    test_suite='nose.collector',
)



'''
This is an implementation of the observer pattern.  It uses function
decorators to achieve the desired event registration mechanism.

For further reference, see

http://en.wikipedia.org/wiki/Observer_pattern

The idea is based on this thread:
http://stackoverflow.com/questions/1904351/python-observer-pattern-examples-tips

'''

import functools
import inspect

try:
    # use python3 signatures if available
    # this takes care of enforcing the correct signature at call time and 
    # provides the correct default arguments
    from inspect import signature
except ImportError: # pragma: no cover
    # python2 has no support for signatures
    def signature(fn):
        return None

__all__ = ['event']
__version__ = '0.2'


class event(object):
    '''
    This class serves as a utility to decorate a function as an event.

    The following example demonstrates its functionality in an abstract way.
    A class method can be decorated as follows:

    >>> class A(object):
    ...     def __init__(self, name):
    ...         self.name = name
    ...
    ...     @event
    ...     def progress(self, first, second):
    ...         print("Doing something...")

    A.progress is the event.  It is triggered after executing the code in the
    decorated progress routine.

    Now that we have a class with some event, let's create an event handler.

    >>> def handler(self, first, second):
    ...     print("%s %s and %s!" % (first, self.name, second))

    Note that the handler (and signal calls) must have the signature defined
    by the decorated event method.

    This handler only greets the object that triggered the event by using its
    name attribute.  Let's create some instances of A and register our new
    event handler to their progress event.

    >>> a = A("Foo")
    >>> b = A("Bar")
    >>> a.progress += handler
    >>> b.progress += handler

    Now everything has been setup.  When we call the method, the event will be
    triggered:

    >>> a.progress("Hello", "World")
    Doing something...
    Hello Foo and World!
    >>> b.progress(second="Others", first="Hi")
    Doing something...
    Hi Bar and Others!

    What happens if we disobey the call signature?

    >>> c = A("World")
    >>> c.progress(second="World")  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    TypeError: progress() missing 1 required positional argument: 'first'

    Class based access is possible as well:

    >>> A.progress(a, "Hello", "Y")
    Doing something...
    Hello Foo and Y!

    Bound methods keep the instance alive:

    >>> f = a.progress
    >>> import weakref, gc
    >>> wr = weakref.ref(a)
    >>> del a
    >>> c=gc.collect()
    >>> assert wr() is not None
    >>> f("Hi", "Z")
    Doing something...
    Hi Foo and Z!

    If we delete the hard reference to the bound method and run the garbage
    collector (to make sure it is run at all), the object will be gone:

    >>> del f
    >>> c=gc.collect()
    >>> assert wr() is None

    '''

    def __init__(self, function):
        '''
        Constructor.

        * function -- The function to be wrapped by the decorator.

        '''
        # Copy docstring and other attributes from function
        functools.update_wrapper(self, function)
        self.__signature__ = signature(function)
        # Used to enforce call signature even when no slot is connected.
        # Can also execute code (called before handlers)
        self.__function = function

    def __set__(self, instance, value):
        '''
        This is a NOP preventing that a boundevent instance is stored.

        This prevents  operations like  `a.progress += handler`  to have
        side effects that result in a cyclic reference.

        http://stackoverflow.com/questions/18287336/memory-leak-when-invoking-iadd-via-get-without-using-temporary

        '''
        pass

    def __get__(self, instance, owner):
        '''
        Overloaded __get__ method.  Defines the object resulting from
        a method/function decorated with @event.

        See http://docs.python.org/reference/datamodel.html?highlight=__get__#object.__get__
        for a detailed explanation of what this special method usually does.

        * instance -- The instance of event invoked.
        * owner -- The owner class.

        '''
        # this case corresponds to access via the owner class:
        if instance is None:
            @functools.wraps(self.__function)
            def wrapper(instance, *args, **kwargs):
                return self.__get__(instance, owner)(*args, **kwargs)
        else:
            wrapper = functools.wraps(self.__function)(boundevent(instance, self.__function))
        wrapper.__signature__ = self.__signature__
        return wrapper


class boundevent(object):
    '''Private helper class for event system.'''

    def __init__(self, instance, function):
        '''
        Constructor.

        * instance -- the instance whose member the event is

        '''
        self.instance = instance
        self.__function = function
        self.__key = ' ' + function.__name__

    @property
    def __event_handlers(self):
        if self.__key not in self.instance.__dict__:
            self.instance.__dict__[self.__key] = []
        return self.instance.__dict__[self.__key]

    def __iadd__(self, function):
        '''
        Overloaded += operator.  It registers event handlers to the event.

        * function -- The right-hand-side argument of the operator; this is the
            event handling function that registers to the event.

        '''
        # Add the function as a new event handler
        self.__event_handlers.append(function)
        # Return the boundevent instance itself for coherent syntax behaviour
        return self

    def __isub__(self, function):
        '''
        Overloaded -= operator.  It removes registered event handlers from
        the event.

        * function -- The right-hand-side argument of the operator; this is the
            function that needs to be removed from the list of event handlers.

        '''
        # Remove the function from the list of registered event handlers
        self.__event_handlers.remove(function)
        # Return the boundevent instance itself for coherent syntax behaviour
        return self

    def __call__(self, *args, **kwargs):
        '''
        Overloaded call method; it defines the behaviour of boundevent().
        When the event is called, all registered event handlers are called.

        * *args -- Arguments given to the event handlers.
        * **kwargs -- Keyword arguments given to the event handlers.

        '''
        # Enforce signature and possibly execute entry code. This makes sure
        # any inconsistent call will be caught immediately, independent of
        # connected handlers.
        result = self.__function(self.instance, *args, **kwargs)
        # Call all registered event handlers
        for f in self.__event_handlers[:]:
            f(self.instance, *args, **kwargs)
        return result
